/* AlnLoader.java
 * ==============
 * Mike Jahr, 8/16/99
 * Machine Translation group, WS99
 * Center for Language and Speech Processing
 */

import java.net.URL;
import java.io.*;
import javax.swing.SwingUtilities;

/**
 * Thread class responsible for loading an alignment file.
 * <code>AlnLoader</code> knows the ".aln" format and, given the url of a
 * file, can load it into a given translation object. It reports any
 * errors it encounters via the {@link Cairo#ReportError(String)
 * Cairo.ReportError} method.
 * @author Mike Jahr
 * @see Cairo
 * @see Translation
 */
public class AlnLoader extends Thread {
    
    //    public static String FILE_SUFFIX = ".aln";
    private Cairo c;
    private Translation trans;
    private URL url;
    
    /**
     * Creates a loader with the given specifications.
     * @see Cairo#openURL(URL)
     */
    public AlnLoader(Cairo inc, Translation intrans, URL inurl) {
	super();
	c = inc;
	trans = intrans;
	url = inurl;
    }
    
    /**
     * Attempts to open an alignment file specified by the URL in the
     * constructor into the translation object.
     */ 
    public void run() {
	//DBG*/c.ReportError("started run");
	
	// if there's already a load thread running, we must wait for it.
	try {
	    Thread t = c.getLoadThread();
	    if (t != null && t.isAlive()) {
		t.interrupt();
		t.join();
	    }
	} catch (InterruptedException ie) {
	    return;
	}
	
	c.setLoadThread(this);
	
	StreamTokenizer st = null;
	try {
	    st = makeTokenizer(null);
	} catch (IOException ioe) {
	    c.ReportError("I/O error with alignment file: " + ioe.getMessage());
	    c.setLoadThread(null);
	    return;
	}
	//DBG*/c.ReportError("made tokenizer");	
	try {
	    int curtag = -1;
	    
	    trans.clear();
	    
	    while (!isInterrupted()) {
		try {
		    curtag = nextTag(st, curtag);
		    //DBG*/System.err.println("got tag " + curtag + ":" + openTag(curtag));
		    
		    if (st.ttype == st.TT_EOF)
			break;
		    
		    switch (curtag) {
		    case HEADER:
			if (trans.encoding == null) {
			    //boolean first = (trans.encoding == null);
			    trans.readHeader(st);
			    
			    if (trans.encoding != null)
				st = makeTokenizer(trans.encoding);
			}
			break;       // "<header>"
		    case SOURCE_SENTENCE:
			trans.source.readWords(st);
			break;       // "<source_sentence>"
		    case TARGET_SENTENCE:
			trans.target.readWords(st);
			break;       // "<target_sentence>";
		    case ALIGNMENT:
			trans.web.readAlignment(st);
			break;       // "<alignment>";
		    case ASCII:
			break;       // "<ascii>";
		    case PROBABILITY:
			trans.target.readProb(st);
			break;       // "<probability>";
		    case NULL_FERTILITY:
			trans.NULL_WORD.readProb(st);
			break;       // "<null_fertility>";
		    case FERTILITIES:
			trans.target.readFertility(st, true);
			break;       // "<fertilities>";
		    case ALT_FERTILITIES:
			trans.target.readFertility(st, false);
			break;       // "<alt_fertilities>";
		    case DISTORTIONS:
			trans.target.readDistortions(st);
			break;       // "<distortions>";
		    case TRANSLATIONS:
			trans.target.readTranslations(st, true, trans.source);
			break;       // "<translations>";
		    case ALT_TRANSLATIONS:
			trans.target.readTranslations(st, false, trans.source);
			break;       // "<alt_translations>";
		    case INVERSE_TRANSLATIONS:
			trans.source.readTranslations(st, true, trans.target);
			break;       // "<inverse_translations>";
		    case ALT_INVERSE_TRANSLATIONS:
			trans.source.readTranslations(st, false, trans.target);
			break;       // "<alt_inverse_translations>";
		    case LANGUAGE_MODEL:
			trans.lm.readNgrams(st, trans.target);
			break;       // "<language_model>";
		    case ALT_LANGUAGE_MODEL:
			trans.lm.readNgrams(st, trans.target);
			break;       // "<alt_language_model>";
		    case REFERENCE_SENTENCE:
			trans.human.readWords(st);
			break;       // "<human_sentence>";
		    case REFERENCE_ALIGNMENT:
			trans.web.readHumanAlignment(st);
			break;       // "<human_alignment>";
		    }
		    
		    if (st.sval != null && st.sval.charAt(2) == '/')
			st.pushBack();
		} catch (FormatException fe) {
		    c.ReportError("Format error in file: " + fe.getMessage());
		    fe.printStackTrace();
		    continue;
		} catch (RuntimeException rte) {
		    c.ReportError(rte.toString());
		    rte.printStackTrace();
		}
	    }
	} catch (IOException ioe) {
	    c.ReportError(ioe.toString());
	}
	
	if (!isInterrupted()) {
	    trans.finish();
	    
	    final Cairo fc = c;
	    SwingUtilities.invokeLater(new Thread() {
		public void run() {
		    fc.validate();
		    fc.repaint();
		}
	    });
	}
	c.setLoadThread(null);

	//DBG*/c.ReportError(trans.source + "\n" + trans.target + "\n" + trans.human);
	//DBG*/c.ReportError("exited run normally");
    }
    
    private int nextTag(StreamTokenizer st, int prevtag) 
	throws IOException, FormatException {
	
	int tok;
	String tag;
	boolean closed = (prevtag != -1);
	
	st.eolIsSignificant(false);
	
	while (true) {
	    tok = st.nextToken();
	    //DBG*/c.ReportError("got token: " + st.sval);
	    if (tok == st.TT_EOF)
		return -1;
	    if (tok == st.TT_WORD && st.sval.charAt(0) == '<') {
		tag = st.sval.toLowerCase();
		if (tag.charAt(1) == '/') {
		    if (prevtag != -1 && !tag.equals(closeTag(prevtag)))
			c.ReportError("Misplaced tag: " + tag);
		    closed = true;
		    continue;
		}
		
		st.eolIsSignificant(true);
		
		if (!closed && prevtag != -1)
		    c.ReportError("Missing tag: " + closeTag(prevtag));
		
		for (int i = 0; i <= LAST_TAG; i++)
		    if (tag.equals(openTag(i)))
			return i;
		throw new FormatException("Unknown tag: " + tag + ".");
	    }
	}
  	//throw new FormatException("Unable to find any tags in file " + fname + ".");
    }

    private StreamTokenizer makeTokenizer(String encoding)
	throws IOException {
	StreamTokenizer st;
	
	//InputStream is = new URL(getDocumentBase(), file).openStream();
	InputStreamReader isr = null;
	
	
	if (encoding != null) {
	    try {
		isr = new InputStreamReader(url.openStream(), encoding);
	    } catch (UnsupportedEncodingException e) {
		c.ReportError("Error in alignment file: character encoding " +
			      trans.encoding + "not supported.");
	    }
	}
    
	if (isr == null)
	    isr = new InputStreamReader(url.openStream());
	
	st = new StreamTokenizer(new BufferedReader(isr));
	st.resetSyntax();
	st.wordChars(0,255);
	st.whitespaceChars('\t','\t');
	st.whitespaceChars(' ', ' ');
	st.whitespaceChars('\n','\n');
	st.whitespaceChars('\r','\r');
	
	st.eolIsSignificant(false);
	st.slashSlashComments(true);
	st.slashStarComments(true);

	return st;
    }
    
    private static String openTag(int i) {
	switch (i) {
	case HEADER:
	    return "<header>";
	case SOURCE_SENTENCE:
	    return "<source_sentence>";
	case TARGET_SENTENCE:
	    return "<target_sentence>";
	case ALIGNMENT:
	    return "<alignment>";
	case ASCII:
	    return "<ascii>";
	case PROBABILITY:
	    return "<probability>";
	case NULL_FERTILITY:
	    return "<null_fertility>";
	case FERTILITIES:
	    return "<fertilities>";
	case ALT_FERTILITIES:
	    return "<alt_fertilities>";
	case DISTORTIONS:
	    return "<distortions>";
	case TRANSLATIONS:
	    return "<translations>";
	case ALT_TRANSLATIONS:
	    return "<alt_translations>";
	case INVERSE_TRANSLATIONS:
	    return "<inverse_translations>";
	case ALT_INVERSE_TRANSLATIONS:
	    return "<alt_inverse_translations>";
	case LANGUAGE_MODEL:
	    return "<language_model>";
	case ALT_LANGUAGE_MODEL:
	    return "<alt_language_model>";
	case REFERENCE_SENTENCE:
	    return "<reference_sentence>";
	case REFERENCE_ALIGNMENT:
	    return "<reference_alignment>";
	default:
	    return null;
	}
    }
    
    private static String closeTag(int i) {
	switch (i) {
	case HEADER:
	    return "</header>";
	case SOURCE_SENTENCE:
	    return "</source_sentence>";
	case TARGET_SENTENCE:
	    return "</target_sentence>";
	case ALIGNMENT:
	    return "</alignment>";
	case ASCII:
	    return "</ascii>";
	case PROBABILITY:
	    return "</probability>";
	case NULL_FERTILITY:
	    return "</null_fertility>";
	case FERTILITIES:
	    return "</fertilities>";
	case ALT_FERTILITIES:
	    return "</alt_fertilities>";
	case DISTORTIONS:
	    return "</distortions>";
	case TRANSLATIONS:
	    return "</translations>";
	case ALT_TRANSLATIONS:
	    return "</alt_translations>";
	case INVERSE_TRANSLATIONS:
	    return "</inverse_translations>";
	case ALT_INVERSE_TRANSLATIONS:
	    return "</alt_inverse_translations>";
	case LANGUAGE_MODEL:
	    return "</language_model>";
	case ALT_LANGUAGE_MODEL:
	    return "</alt_language_model>";
	case REFERENCE_SENTENCE:
	    return "</reference_sentence>";
	case REFERENCE_ALIGNMENT:
	    return "</reference_alignment>";
	default:
	    return null;
	}
    }
    
    // these are tag constants for internal use
    private static final int HEADER = 0;
    private static final int SOURCE_SENTENCE = 1;
    private static final int TARGET_SENTENCE = 2;
    private static final int ALIGNMENT = 3;
    private static final int ASCII = 4;
    private static final int PROBABILITY = 5;
    private static final int NULL_FERTILITY = 6;
    private static final int FERTILITIES = 7;
    private static final int ALT_FERTILITIES = 8;
    private static final int DISTORTIONS = 9;
    private static final int TRANSLATIONS = 10;
    private static final int ALT_TRANSLATIONS = 11;
    private static final int INVERSE_TRANSLATIONS = 12;
    private static final int ALT_INVERSE_TRANSLATIONS = 13;
    private static final int LANGUAGE_MODEL = 14;
    private static final int ALT_LANGUAGE_MODEL = 15;
    private static final int REFERENCE_SENTENCE = 16;
    private static final int REFERENCE_ALIGNMENT = 17;
    private static final int LAST_TAG = REFERENCE_ALIGNMENT;
}
