/* Cairo.java
 * ==========
 * Mike Jahr, 7/13/99
 * Machine Translation group, WS99
 * Center for Language and Speech Processing
 * 
 * Tool for examining alignment files generated by statistical translator.
 */

import javax.swing.*;
import java.awt.Container;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.event.*;
import java.applet.Applet;
import java.io.File;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.StringTokenizer;

/**
 * The <code>Cairo</code> object lies at the heart of the package: it
 * manages the top-level layout of the GUI and keeps track of all other
 * objects.
 * The <code>main()</code> method is implemented here so Cairo is
 * executable as a standalone application; it creates and executes a
 * CairoApp. In order to embed Cairo in a web page, embed a
 * <code>CairoApplet</code> object instead.
 * @author Mike Jahr
 * @version 0.95
 * @see CairoApp
 * @see CairoApplet
 * @see CairoWrap
 */
public class Cairo extends JSplitPane {
    /** Version information. */
    public static final String VERSION = "Cairo v0.95";
    /** Author information. */
    public static final String BYLINE = "Mike Jahr and Noah Smith";
    
    /** Wrapper for {@link CairoApp#main(String[]) CairoApp.main}. */
    public static void main(String args[]) {
	CairoApp.main(args);
    }
    
    /////////////////////// CONSTANTS //////////////////////
    
    /**
     * The default orientation of the sentences.
     * For horizontal alignment use BoxLayout.X_AXIS, for vertical
     * alignment use BoxLayout.Y_AXIS.
     */
    public static final int DEFAULT_SENTENCE_AXIS = BoxLayout.X_AXIS;
    /** Default width of Cairo window; overridden in applet mode. */
    public static final int DEFAULT_WIDTH = 800;
    /** Default height of Cairo window; overridden in applet mode. */
    public static final int MIN_HEIGHT = 100;
    /** Default width of Cairo window; overridden in applet mode. */
    public static final int MIN_WIDTH = 200;
    /** Default height of Cairo window; overridden in applet mode. */
    public static final int DEFAULT_HEIGHT = 600;
    /** Default directory to search for alignment files. */
    public static final String DEFAULT_ALIGN_DIR = "alignments/";
    /** Suffix for alignment files. */
    public static final String ALIGN_SUFFIX = ".aln";
    /**
     * Name of index file to use to list directory contents when
     * running in applet mode.
     * @see FileMenuApplet
     */
    public static final String INDEX_FILE_NAME = "index";
    
    private static final int INFO_WIDTH = 150;
    private static final int SENT_OUTER_MARGIN = 2;
    private static final int SENT_INNER_MARGIN = 0;
    private static final Dimension minSize = 
	new Dimension(MIN_WIDTH, MIN_HEIGHT);
    private static final Dimension prefSize = 
	new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT);
    private static final double MAIN_SPLIT = 0.4;
    private static final double INFO_SPLIT = 0.2;

    ////////////////// INSTANCE VARIABLES //////////////////
    
    private CairoWrap wrap = null;
    private Container base = null;
    private Thread loadThread = null;
    //private InputStream instream = null;
    private String errmsg = null;
    private Translation trans = null;
    private JSplitPane infoPane = null;
    private Container sentencePane = null;
    private JSplitPane mainPane = null;
    private OptionMenu oMenu = null;
    private ProbTable tt, dt, lt, ft;
    private boolean finished = false;
    
    //////////////////// APPLET METHODS ////////////////////
    
//     /** 
//      * Creates and initializes a standalone Cairo application.
//      */
//     public Cairo() {
// 	this(null);
// 	addWindowListener(new AppCloser());
//     }
    /**
     * Creates and initializes a Cairo using the given CairoWrap.
     * Pass a CairoApplet to make an applet, or a CairoApp to
     * make a standalone application.
     * @see CairoApplet
     */
    public Cairo(CairoWrap incw) {
	super(((DEFAULT_SENTENCE_AXIS == BoxLayout.X_AXIS) 
	       ? JSplitPane.VERTICAL_SPLIT : 
	       JSplitPane.HORIZONTAL_SPLIT), 
	      true, new JPanel(), new JPanel());
	if (incw == null)
	    throw new Error("Attempt to create Cairo with null CairoWrap");
	mainPane = this;
	wrap = incw;
	base = wrap.getContentPane();
    }
    /**
     * Interrupts any threads which are still alive and performs any
     * necessary cleanup.
     */
    public void stop() {
	if (loadThread != null && loadThread.isAlive())
	    loadThread.interrupt();
	loadThread = null;
    }

    ////////////////// INITIALIZATION METHODS //////////////////

    /**
     * Initializes the Cairo object and all its helping objects.
     * Sets up the gui, creates menus, and initializes translation 
     * and sentence objects.
     * @see Translation
     * @see Sentence
     * @see #createMenuBar()
     */
    protected void init() {
	if (finished == true) return;
	
	wrap.getRootPane().setJMenuBar(createMenuBar());
	mainPane.setOneTouchExpandable(true);
	base.add(mainPane);
// 	mainPane.setDividerLocation
// 	    ((int) (((DEFAULT_SENTENCE_AXIS == BoxLayout.X_AXIS)
// 		     ? base.getHeight() : base.getWidth()) * 0.3));
// 	//wrap.getRootPane().validate();
	trans = new Translation(this, DEFAULT_SENTENCE_AXIS);
	
	createInfoPane(DEFAULT_SENTENCE_AXIS);
	createSentencePane(DEFAULT_SENTENCE_AXIS);
// 	mainPane.setDividerLocation
// 	    ((int) (((DEFAULT_SENTENCE_AXIS == BoxLayout.X_AXIS)
// 		     ? base.getHeight() : base.getWidth()) * 0.3));

	sizePanes(DEFAULT_SENTENCE_AXIS);
	finished = true;
    }

    /** Sets pane sizes to reasonable defaults. */
    private void sizePanes(int axis) {
	boolean horiz = (axis == BoxLayout.X_AXIS);
	mainPane.setDividerLocation
	    ((int) ((horiz ? base.getHeight() : base.getWidth()) * MAIN_SPLIT));
	infoPane.setDividerLocation
	    ((int) ((horiz ? base.getWidth() : base.getHeight()) * INFO_SPLIT));
	try { ((JSplitPane)
	       infoPane.getRightComponent()).setDividerLocation
		  ((int) ((horiz ? base.getHeight() : base.getWidth()) *
			  (1-MAIN_SPLIT) / 2) - 10);
	} catch (Exception ignored) {}
    }
    /**
     * Creates the menu bar and add all menus.
     * Uses {@link CairoWrap#createFileMenu(Cairo) CairoWrap.createFileMenu}
     * to get the relevent FileMenu.
     * @see FileMenuApp
     * @see FileMenuApplet
     * @see OptionMenu
     * @see HelpMenu
     */
    protected JMenuBar createMenuBar() {
	JMenuItem item;
	
	JMenuBar mb = new JMenuBar();
	
	mb.add(wrap.createFileMenu(this));
	oMenu = new OptionMenu(this);
	mb.add(oMenu);
	HelpMenu hm = new HelpMenu(this);
	mb.add(hm);
	// not yet implemented by swing
	// mb.setHelpMenu(hm);
	return mb;
    }
    /**
     * Creates the infoPane panel; this contains the {@link Translation
     * Translation} object and displays general sentence information.
     */
    private void createInfoPane(int axis) {
	boolean horiz = (axis == BoxLayout.X_AXIS);
	Box head = new Box(horiz ? BoxLayout.Y_AXIS : BoxLayout.X_AXIS);
	//JScrollPane head = new JScrollPane(trans);
	//head.add(head.createVerticalGlue());
	//head.setColumnHeaderView(jl);
	JLabel jl = new JLabel("Sentence Stats");
	jl.setAlignmentX(CENTER_ALIGNMENT);

	head.add(head.createVerticalStrut(5));
	head.add(jl);
	head.add(head.createVerticalStrut(5));
	head.add(new JScrollPane(trans));
	head.add(head.createVerticalGlue());
	//head.add(jl);
	//head.add(new JScrollPane(trans));

	tt = new ProbTable();
	dt = new ProbTable();
	lt = new ProbTable();
	ft = new ProbTable();

	updateTables(null);
	
	infoPane = new JSplitPane((horiz ? JSplitPane.HORIZONTAL_SPLIT :
				   JSplitPane.VERTICAL_SPLIT),
				  true, head, new JPanel());
	infoPane.setOneTouchExpandable(true);
	//infoPane.setDividerLocation(INFO_WIDTH);
	mainPane.setBottomComponent(infoPane);	
	//mainPane.validate();
	//infoPane.setDividerLocation(0.2); /*(int) (horiz ? infoPane.getWidth() :
	//base.getHeight()) / 5);*/
	showFourUp();
	//((JSplitPane) infoPane.getRightComponent()).setDividerLocation(0.5);
    }
    
    /**
     * Creates and formats the sentencePane panel; this contains the
     * {@link Web Web} and {@link Sentence Sentence} objects.
     */
    private void createSentencePane(int axis) {
	//sentAxis = axis;
	
	boolean horiz = (axis == BoxLayout.X_AXIS);
	//Container c = getContentPane();
	Box b = new Box(horiz ? BoxLayout.Y_AXIS : BoxLayout.X_AXIS);
	
	trans.setAxis(axis);
	
	b.add(horiz ? b.createVerticalStrut(SENT_OUTER_MARGIN) :
	      b.createHorizontalStrut(SENT_OUTER_MARGIN));
	b.add(trans.source);

	b.add(horiz ? b.createVerticalStrut(SENT_INNER_MARGIN) :
	      b.createHorizontalStrut(SENT_INNER_MARGIN));
	b.add(trans.web);
	b.add(horiz ? b.createVerticalStrut(SENT_INNER_MARGIN) :
	      b.createHorizontalStrut(SENT_INNER_MARGIN));
	
	b.add(trans.web.getActiveSentence());
	
	if (trans.web.isAltShown())
	    b.add(trans.web.getAltSentence());
	
	Component c = horiz ? b.createVerticalStrut(SENT_OUTER_MARGIN) :
	    b.createHorizontalStrut(SENT_OUTER_MARGIN);
	b.add(c);
	trans.web.setMargin(c);
	
	//JScrollPane top = new JScrollPane(b);
	//DBG*/ReportError("created sentences");
	mainPane.setTopComponent(new JScrollPane(b));
	if (finished)
	    mainPane.setOrientation((horiz ? JSplitPane.VERTICAL_SPLIT : 
				     JSplitPane.HORIZONTAL_SPLIT));
	//mainPane.resetToPreferredSizes();
	/*
	mainPane.setDividerLocation
	    ((int) (((DEFAULT_SENTENCE_AXIS == BoxLayout.X_AXIS)
		     ? base.getHeight() : base.getWidth()) * 0.3));
	*/
	sentencePane = b;
    }

    ////////////////// ACCESSORS //////////////////

    /**
     * Returns the current {@link Translation Translation} object.
     */
    public Translation getTrans() {
	return trans;
    }
    /**
     * Required by the {@link CairoWrap CairoWrap} interface;
     * returns a {@link FileMenuApp FileMenuApp}.
     */
    public JMenu createFileMenu(Cairo inc) {
	return new FileMenuApp(inc);
    }
    /**
     * Required by the {@link CairoWrap CairoWrap} interface;
     * returns <code>this</code>.
     */
    public Cairo getCairo() {
	return this;
    }
    /**
     * Returns a thread which is currently loading an alignment file, if
     * any is currently executing.
     */
    public Thread getLoadThread() {
	return loadThread;
    }
    /**
     * Updates the currently loading thread to reflect the given value.
     */
    public void setLoadThread(Thread inthread) {
	loadThread = inthread;
    }
    /**
     * Overrides JComponent.getMinimumSize().
     * @see #getPreferredSize()
     */
    public Dimension getMinimumSize() {
	return minSize;
    }
    /**
     * Overrides JComponent.getPreferredSize().
     * @see #getMinimumSize()
     */
    public Dimension getPreferredSize() {
	return prefSize;
    }

    ////////////////// GUI METHODS //////////////////
    
    /**
     * Updates the probability tables using the given word.
     * If <code>w == null</code> it clears the tables.
     * @see ProbTable
     */
    public void updateTables(Word w) {
	if (w == null) {
	    //DBG*/System.err.println("Updating tables with null");
	    tt.setModel(Word.emptyTransTM);
	    dt.setModel(Word.emptyDistTM);
	    ft.setModel(Word.emptyFertTM);
	    lt.setModel(LanguageModel.emptyLMTM);
	} else {
	    //DBG*/System.err.println("Updating tables with word index=" + w.index);
	    tt.setModel(w.getTranslations());
	    dt.setModel(w.getDistortions());
	    ft.setModel(w.getFertilities());
	    try {
		lt.setModel(trans.lm.getNgrams(((TargetWord)w).index()));
	    } catch (ClassCastException cce) {
		lt.setModel(LanguageModel.emptyLMTM);
	    }
	}
	tt.sizeColumnsToFit(-1);
	dt.sizeColumnsToFit(-1);
	ft.sizeColumnsToFit(-1);
	lt.sizeColumnsToFit(-1);
    }
    
    // stream methods. cairo acts as an intermediary between web and
    // optionmenu.
    /**
     * Updates the streams menu to reflect the given string of streams by
     * calling {@link OptionMenu#updateStreams(String)
     * OptionMenu.updateStreams}.
     */
    public void updateStreams(String s) {
	oMenu.updateStreams(s);
    }
    
    /**
     * Shows or hides the given stream by calling
     * {@link Translation#setStreamVisible(int, boolean)
     * Translation.setStreamVisible}.
     */
    public void setStreamVisible(int s, boolean vis) {
	trans.setStreamVisible(s, vis);
    }
    
    /** Displays all four tables simultaneously. */
    public void showFourUp() {
	Box top = new Box(BoxLayout.X_AXIS);
	top.add(top.createHorizontalGlue());
	top.add(new JScrollPane(tt));
	top.add(top.createHorizontalGlue());
	top.add(new JScrollPane(ft));
	top.add(top.createHorizontalGlue());

	Box bot = new Box(BoxLayout.X_AXIS);
	bot.add(bot.createHorizontalGlue());
	bot.add(new JScrollPane(dt));
	bot.add(bot.createHorizontalGlue());
	bot.add(new JScrollPane(lt));
	bot.add(bot.createHorizontalGlue());
	
	JSplitPane tables = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
					   true, top, bot);
	infoPane.setRightComponent(tables);
	tables.setDividerLocation(infoPane.getHeight()/2); /*(int) 
					  ((infoPane.getHeight() 
				  - infoPane.getDividerSize()) / 2));*/
	//System.err.println("infoPane dimensions: " + infoPane.getBounds());
	
	if (finished) infoPane.validate();
    }
    /** 
     * Displays two tables at a time. Users can tab between translations
     * and distortions, and n-grams and fertilities.
     */
    public void showTwoUp() {
	JTabbedPane left = new JTabbedPane();
	JTabbedPane right = new JTabbedPane();
	left.addTab("Translations", new JScrollPane(tt));
	left.addTab("Distortions", new JScrollPane(dt));
	right.addTab("N-grams", new JScrollPane(lt));
	right.addTab("Fertilities", new JScrollPane(ft));
	
	JSplitPane tables = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
					   true, left, right);
	tables.setDividerLocation((int) (infoPane.getWidth() / 2));
	
	infoPane.setRightComponent(tables);
	if (finished) infoPane.validate();
    }
    /** 
     * Displays tables one at a time; users select which table to view
     * via a tabbed interface.
     */
    public void showOneUp() {
	JTabbedPane tables = new JTabbedPane();

	tables.addTab("Translations", new JScrollPane(tt));
	tables.addTab("Distortions", new JScrollPane(dt));
	tables.addTab("N-grams", new JScrollPane(lt));
	tables.addTab("Fertilities", new JScrollPane(ft));
	
	infoPane.setRightComponent(tables);
	if (finished) infoPane.validate();
    }
    /**
     * Displays sentences vertically on the left side of the window, with 
     * the tables on the right.
     */
    public void showVert() {
	if (trans.web.getAxis() == BoxLayout.Y_AXIS)
	    return;
	createSentencePane(BoxLayout.Y_AXIS);
	infoPane.setOrientation(JSplitPane.VERTICAL_SPLIT);
	sizePanes(BoxLayout.Y_AXIS);
	validate();
    }

    /**
     * Displays sentences horizontally at the top of the window, with 
     * the tables below.
     */
    public void showHoriz() {
	if (trans.web.getAxis() == BoxLayout.X_AXIS)
	    return;
	createSentencePane(BoxLayout.X_AXIS);
	infoPane.setOrientation(JSplitPane.HORIZONTAL_SPLIT);
	sizePanes(BoxLayout.X_AXIS);
	validate();
    }
    
    ////////////////////// I/O METHODS //////////////////////
    
    /**
     * Attempts to open an alignment file at the url specified by the
     * given string. If the argument does not represent a valid 
     * absolute url, displays an error and does nothing. Otherwise, it
     * spawns a loader thread.
     * @see #openURL(URL)
     * @see #openURL(URL, String)
     * @see #openFile(File)
     */
    public synchronized void openURL(String inurl) {
	openURL(null, inurl);
    }
    
    /**
     * Attempts to open an alignment file at the url specified by the
     * given string, with the given url as the base directory. If the 
     * arguments do not represent a valid url, displays an error and 
     * does nothing. Otherwise, it spawns a loader thread.
     * @see #openURL(URL)
     * @see #openURL(String)
     * @see #openFile(File)
     */
    public synchronized void openURL(URL base, String inurl) {
	//fname = infile;
	try {
	    if (base == null)
		openURL(new URL(inurl));
	    else
		openURL(new URL(base, inurl));
	} catch (MalformedURLException e) {
	    ReportError("Error with url <" + inurl + ">: " 
			+ e.getMessage());
	}
    }
    /**
     * Attempts to load the alignment file at the given url by spawning a
     * loader thread. If the argument does not represent a valid 
     * url, displays an error and does nothing.
     * @see #openURL(URL)
     * @see #openURL(String)
     * @see #openURL(URL, String)
     */
    public synchronized void openURL(URL url) {
	new AlnLoader(this, trans, url).start();
	updateTables(null);
    }
    /**
     * Attempts to load the given alignment file by spawning a
     * loader thread. If the argument does not represent a valid 
     * file, displays an error and does nothing.
     * @see #openURL(URL)
     * @see #openURL(String)
     * @see #openURL(URL, String)
     */
    public synchronized void openFile(File infile) {
	try {
	    openURL(infile.toURL());
	} catch (MalformedURLException e) {
	    ReportError("Error opening file " + infile + ": " 
			+ e.getMessage());
	}
	/*
	try {
	    instream = new FileInputStream(infile);
	    newThread = new AlnLoader(this, trans);
	    newThread.start();
	    updateTables(null);
	} catch (IOException ioe) {
	    ReportError("Error with file " + infile.getAbsolutePath() 
			+ ": " + ioe.getMessage());
	}
	*/
    }
    
    ///////////////////// UTILITY METHODS /////////////////////

    /**
     * Displays an error dialog with the given string.
     */
    public void ReportError(String err) {
	//DBG*/System.err.println(err);
	JTextArea jte = new JTextArea(err, 6, 40);
	jte.setLineWrap(true);
	jte.setWrapStyleWord(true);

	JOptionPane.showOptionDialog
	    (this, new JScrollPane(jte), "Cairo error", 
	     JOptionPane.DEFAULT_OPTION,
	     JOptionPane.WARNING_MESSAGE, null, null, null);
    }
    
    /**
     * Trivial listener class which calls <code>System.exit()</code> when
     * its window is closed.
     * @author Mike Jahr
     */
    protected static final class AppCloser extends WindowAdapter {
        public void windowClosing(WindowEvent e) {
	    System.exit(0);
	}
    }
}   // end class Cairo
