/* FileMenuApplet.java
 * ===================
 * Mike Jahr, 8/13/99
 * Machine Translation group, WS99
 * Center for Language and Speech Processing
 * 
 */

import javax.swing.*;
import java.awt.event.*;
import java.net.URL;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;

/**
 * File menu class for the Cairo applet. 
 * Because applets cannot view the contents of directories, it relies on
 * a file called "index" in the base directory for a complete listing. It
 * uses an IndexLoader to load the file.
 * The menu consists of a hierarchical "Open" menu with all files listed
 * in the index. There is also  an "Open other" menu item which allows users 
 * to open files not in the index, and a "Reload index" to update the
 * hierarchical menu.
 * @author Mike Jahr
 * @see CairoApplet
 * @see Cairo
 * @see IndexLoader
 * @see FileMenuApp
 */
public class FileMenuApplet extends JMenu implements ActionListener {
    private CairoApplet master;
    private String[] alignDirs = null;
    
    /**
     * Creates a <code>FileMenuApplet</code> object for the given
     * CairoApplet, which indexes all directories in the colon-delimited
     * list <code>dirs</code>. It spawns an <code>IndexLoader</code>
     * thread to load the index file.
     * @see IndexLoader
     */
    public FileMenuApplet(CairoApplet ca, String dirs) {
	super("File");
	master = ca;
	
	JMenuItem item;
	//file.setMnemonic(KeyEvent.VK_F);
	
	item = new JMenu("Open");
	item.setEnabled(false);
	add(item);
	new IndexLoader(this, master.getCairo().DEFAULT_ALIGN_DIR).start();
	
	item = new JMenuItem(FILE_OPEN); //, KeyEvent.VK_O);
	item.addActionListener(this);
	item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, 
						   ActionEvent.ALT_MASK));
	add(item);
	
	addSeparator();
	
	item = new JMenuItem(FILE_RELOAD); //, KeyEvent.VK_O);
	item.addActionListener(this);
	add(item);
	
	/*
	  item = new JMenuItem(FILE_CLOSE); //, KeyEvent.VK_W);
	  item.addActionListener(this);
	  item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W, 
	  ActionEvent.ALT_MASK));
	  add(item);
	*/
    }
    
    /** Required by the ActionListener interface; responds to menu events.
     */
    public void actionPerformed(ActionEvent ae) {
	String result;
	JMenuItem source = (JMenuItem)(ae.getSource());
	
	if (source.getText().equals(FILE_OPEN)) {
	    result = JOptionPane.
		showInputDialog(master, "Please specify the URL of an alignment " 
				+ "(.aln) file to open:", "Open", 
				JOptionPane.QUESTION_MESSAGE);
	    if (result != null && result.length() > 0) {
		//if (!result.startsWith(ALIGN_DIR))
		//	result = ALIGN_DIR + result;
		//if (!result.endsWith(ALIGN_SUFFIX))
		//	result = result + ALIGN_SUFFIX;
		master.getCairo().openURL(master.getDocumentBase(), result);
	    }
	} else if (source.getText().equals(FILE_RELOAD)) {
	    new IndexLoader(this, master.getCairo().DEFAULT_ALIGN_DIR).start();
	} else {
	    FileMenuItem fm = null;
	    try { fm = (FileMenuItem) source; } catch (Exception e) {}
	    if (fm != null)
		master.getCairo().openURL(fm.file);
	}
	
	//String s = "Action event detected.\n"
	//	+ "    Event source: " + source.getText();
	//		+ " (an instance of " + getClassName(source) + ")";
	//System.err.println(s);
    }
    
    private static final String FILE_OPEN = "Open other...";
    private static final String FILE_CLOSE = "Close";
    private static final String FILE_OPEN_SUB = "Open";
    private static final String FILE_RELOAD = "Reload index file";
    //private static final String FILE_OPEN = "Open an alignment...";
    
    /**
     * Simple menu item object which stores a URL representing a file.
     * @author Mike Jahr
     */
    protected static class FileMenuItem extends JMenuItem {
	/** The location of the file. */
	public URL file;
	/**
	 * Creates a <code>FileMenuItem</code> object with the given name 
	 * and file attributes.
	*/
	public FileMenuItem(String name, URL infile) {
	    super(name);
	    file = infile;
	}
    }
    /**
     * Loader class which creates a hierarchical Open menu based on an
     * <code>index</code> file.
     * @author Mike Jahr
     * @see FileMenuApplet
     */
    protected static class IndexLoader extends Thread {
	private FileMenuApplet master;
	private Vector menuItems;
	private String title;
	private URL docBase;
	
	/**
	 * Creates an <code>IndexLoader</code> object which will add a
	 * menu with the given name to the given <code>JMenu</code>.
	 */
	public IndexLoader(FileMenuApplet inm, String intitle) {
	    super();
	    master = inm;
	    menuItems = new Vector();
	    title = intitle;
	}
	/** 
	 * Thread method which does al the work. If an error is
	 * encountered while reading the index file, a message is
	 * displayed and a disabled menu item is added instead.
	 */
	public void run() {
	    JMenu ret = null;
	    
	    try {
		StreamTokenizer st;
		URL base = new URL(master.master.getDocumentBase(), 
				   Cairo.DEFAULT_ALIGN_DIR + 
				   Cairo.INDEX_FILE_NAME);
		InputStream is = base.openStream();
		st = new StreamTokenizer
		    (new BufferedReader(new InputStreamReader(is)));
		
		st.resetSyntax();
		st.wordChars(0,255);
		st.whitespaceChars('\n', '\n');
		st.eolIsSignificant(false);
		st.slashSlashComments(false);
		st.slashStarComments(false);
		
		st.nextToken(); // prime st for helper
		ret = readIndex(st, base, "");
	    } catch(Exception e) {
		master.master.getCairo().ReportError("Error reading directory index file: " +
					      e.getMessage());
		e.printStackTrace();
	    }

	    if (ret == null)
		ret = new JMenu(title);
	    if (ret.getItemCount() == 0)
		ret.setEnabled(false);
	    
	    final IndexLoader il = this;
	    final JMenu om = ret;
	    
	    SwingUtilities.invokeLater(new Thread() {
		public void run() {
		    Enumeration e = il.menuItems.elements();
		    while (e.hasMoreElements())
			((JMenuItem) e.nextElement()).addActionListener(il.master);
		    
		    il.master.remove(0);
		    il.master.insert(om, 0);
		}
	    });
	}
	
	// note to noah: don't touch...this works. i changed it
	// so that st reads one line at a time, which seems to be
	// much better. also, now we can mix files and directories
	// at will in the index file and this doesn't balk.
	
	// spiffy recursive function to turn the index file into a
	// heirarchical menu. see inline comments
	private JMenu readIndex(StreamTokenizer st, URL base, String path)
	    throws IOException {
	    
	    int tok, i, len = path.length();
	    FileMenuItem item;
	    JMenu ret = null; //new JMenu(newDir);
	    String newDir, newPath, fileName;
	    
	    //DBG*/System.out.println("starting helper. path is: " + path);
	    //DBG*/System.out.println("next token is: " + st.sval);
	    
	    if (len > 0) {
		i = path.lastIndexOf('/', len-2); // get index of prev '/'
		//if (i < 0) i = 0; // in case we're the first subdirectory
		ret = new JMenu(path.substring(i+1, len-1));
	    } else // we're the root directory
		ret = new JMenu(title);
	    
	    while (true) {
		if (st.ttype == st.TT_EOF) {
		    //DBG*/System.out.println("exiting helper");
		    return ret;
		}
		newPath = st.sval;
		if (newPath.startsWith(path)) { // it involves this directory
		    fileName = newPath.substring(len);
		    i = fileName.indexOf('/');
		    if (i == -1) { // no '/'--this is a file name
			//DBG*/System.out.println("adding item: " + fileName);
			item = new FileMenuItem(fileName, new URL(base,
								  newPath));
			//System.err.println(item.file.toString());
			//item.addActionListener(fma);
			menuItems.add(item);
			ret.add(item);
			st.nextToken();
			//DBG*/System.out.println("next token is: " + st.sval);
		    } else { // there is a '/'--this is a subdirectory
			newDir = path.concat(fileName.substring(0, i+1));
			//DBG*/System.out.println("making subdir: " + newDir);
			ret.add(readIndex(st, base, newDir));
		    }
		} else { // this is in some higher directory
		    return ret;
		}
	    }
	}
    } // end class IndexLoader
} // end class FileMenu
