/*
 * LanguageModel class
 * Noah A. Smith, 13 July 1999
 * Last edited 21 July 1999, NAS
 * Statistical Machine Translation Team at WS '99
 */

import java.util.Vector;
import java.io.StreamTokenizer;
import java.io.IOException;
import java.util.TreeSet;
import javax.swing.table.*;
import java.lang.Class;
import java.lang.Integer;

/** A data structure consisting of a collection of language model (i.e., n-gram) probabilities relevant to the current alignment.
    @author Noah A. Smith
    @see SPTableModel
*/
public class LanguageModel{
    private Vector positions = new Vector();  /* a vector of SPTableModels */
    private int currPos = 0;
    private int gram = 1;
    private Web wb;
    public static SPTableModel emptyLMTM = new SPTableModel("N-gram");

    /** Instantiate a new language model for the target sentence in Web 'inwb'. */
    public LanguageModel(Web inwb) {
	wb = inwb;
    }

    /** Add an n-gram culminating in position 'index' with word 's' with probability 'prob' and correctness 'c'. */
    public void addNgram(int index, String s, Probability prob, 
			 boolean corr){
	StringProb sp = new StringProb(s, prob);
	for(int j = positions.size(); j <= index; j++){
	    positions.addElement(new SPTableModel("N-gram"));
	}
	if(corr) sp.setColor(wb.getAllColor());
	((SPTableModel)positions.elementAt(index)).add(sp);
	//*DEBUG*/ System.err.println("Added:  " + s + "\t" + prob + "\t" + positions.size());
    }

    /** Clear the language model - use only when loading a new translation. */
    public void clear(){
	positions.clear();
    }
    
    /** Change the current position to 'p'. */
    public void setPos(int x){
	if(x < positions.size())
	    currPos = x;
    }
    
    /** Return a vector of the N-grams which apply to the current position. */
    public SPTableModel getCurrNgrams(){
	return (SPTableModel)positions.elementAt(currPos);
    }

    /** Return a vector of the N-grams which apply to the given position 'p'. */
    public SPTableModel getNgrams(int p){
	if(p >= 0 && p < positions.size()){
	    //*DEBUG*/ ((SPTableModel)positions.elementAt(p)).print();
	    return (SPTableModel)positions.elementAt(p);
	}
	else{
	    //*DEBUG*/ System.err.println("Request for n-grams relevant to position out of bounds:  " + p + " - current bounds are  (0, " + positions.size() + ").");
	    return emptyLMTM;
	}
    }
    
    /** Read n-gram information; see documentation on .aln format. */
    /* Format is:
       { <language_model> <newline>
       (<target word position> <word> <p(target word | previous n - 1words in sentence)> <newline>)+
         </language_model> }
	 where {} indicate what is read. */
    public void readNgrams(StreamTokenizer st, Sentence sent) throws FormatException {
	boolean alt = false, done = false;
	int pos = -1;
	String text, temp;
	Probability prob = new Probability();
	boolean corr = false;

	if(st.sval.equals("<language_model>"))
	    alt = false;
	else if(st.sval.equals("<alt_language_model>"))
	    alt = true;
	else
	    throw new FormatException("Corrupt file format!  LM  01");
	try{
	    switch(st.nextToken()){
	    case st.TT_EOF:  throw new FormatException("Unexpected end of file!  LM  02");
	    case st.TT_WORD:  throw new FormatException("Expecting end of line, got string!  LM  06");
	    case st.TT_NUMBER:  throw new FormatException("Zut alors!  LM  04");
	    }
	    while(!done){
		switch(st.nextToken()){
		case st.TT_NUMBER:  throw new FormatException("Zut alors!  LM  04.1");
		case st.TT_EOF:  throw new FormatException("Unexpected end of file!  LM  05");
		case st.TT_EOL:  throw new FormatException("Unexpected end of line!  LM  07");		
		case st.TT_WORD:  
		    if(st.sval.equals("</language_model>"))
			return;
		    else if(st.sval.equals("</alt_language_model>"))
			return;
		    else
			pos = Integer.parseInt(st.sval);		    
		}	
		
		if(pos == -1)
		    throw new FormatException("Expected position that was not found!  LM  07.1");
		text = "";
		if(pos >= (gram - 1) && pos < sent.getLength()){
		    for(int i = gram - 1; i > 0; i--){
			text = text + sent.getWordAt(pos - i).getText() + " ";
		    }
		}
		else if(pos < gram - 1){ 
		    text = "<s> ";
		    for(int i = 0; i < pos; i++){
			text = text + sent.getWordAt(i).getText() + " ";
		    }
		}
		else if(pos == sent.getLength()){
		    for(int i = gram - 1; i > 1; i--){
			text = text + sent.getWordAt(pos - i).getText() + " ";
		    }
		}
		    
		//		System.err.println("SL = " + sent.getLength());
		temp = prob.read(st, false);
		if(temp != null){
		    text = text + temp;
		    if(pos < sent.getLength() && temp.equals(sent.getWordAt(pos).getText())){
			//*DEBUG*/ System.err.println("pos < sent.getLength:  " + pos + "\t\ttemp == word:  " + temp + " == " + sent.getWordAt(pos).getText());
			corr = true;
		    }
		    else if(pos < sent.getLength()){
			//*DEBUG*/ System.err.println("pos < sent.getLength:  " + pos + "\t\ttemp != word:  " + temp + " != " + sent.getWordAt(pos).getText());
			corr = false;
		    }
		    else if(pos == sent.getLength() && temp.equals("</s>")){
			//*DEBUG*/ System.err.println("pos == sent.getLength:  " + pos + "\t\ttemp == </s>:  " + temp + " == </s>");
			corr = true;
		    }
		    if(prob.read(st, corr) != null)
			throw new FormatException("Bad file format!  LM  10.1");
		}
		else{ /* handle correct translations */
		    corr = true;
		    if(pos < sent.getLength())
			text = text + (sent.getWordAt(pos).getText());
		    else if(pos == sent.getLength()){
			text = text + "</s>";
		    }
		    else throw new FormatException("Bad file format:  position in language model values too high!  LM  10.2");
		}
		
		
		//*DEBUG*/ System.err.println("Read n-gram:  " + pos + " " + text + " " + prob.toString() + " " + corr);
		addNgram(pos, text, prob, corr);
		switch(st.nextToken()){
		case st.TT_NUMBER: throw new FormatException("Zut alors!  LM  11");
		case st.TT_EOF:  throw new FormatException("Unexpected end of file!  LM  17");
		case st.TT_EOL:  throw new FormatException("Unexpected end of line!  LM  18");
		case st.TT_WORD:
		    if((st.sval.equals("</language_model>")) || (st.sval.equals("</alt_language_model>")))
			done = true;
		    else
			st.pushBack();
		}
	    }
	}
	catch(IOException ioe){
	    throw new FormatException("Unexpected end of file!  LM  19");
	}
	//*DEBUG*/ System.err.println("Size of positions vector:  " + positions.size());
    }

    /** Set the value of N (e.g., 3 for trigram). */
    public void setGram(int g){
	if(g >= 0)
	    gram = g;
    }
}
