/* Sentence.java
 * =============
 * Mike Jahr, 7/14/99
 * 
 * Last edited 7/21/99 Noah A. Smith
 * Machine Translation group, WS99
 * Center for Language and Speech Processing
 */

import javax.swing.*;
import javax.swing.plaf.ComponentUI;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.io.StreamTokenizer;
import java.io.IOException;

/**
 * Represents a single sentence and all related data; responsible for
 * information storage as well as visual display.
 * Visually, a Sentence is implemented as a subclass of javax.swing.Box,
 * which basically means it is either a row or column of Words. It
 * provides methods for changing the GUI as well as accessors for
 * information about individual words.
 * @author Mike Jahr
 * @see Web
 * @see Word
 */
public class Sentence extends Box { 
    //implements ComponentListener {
    //public static final int HEIGHT = 40;
    /** Constant for marking target sentences. */
    public static final int TARGET = 0;
    /** Constant for marking source sentences. */
    public static final int SOURCE = 1;
    /** Constant for marking human sentences. */
    public static final int HUMAN = 2;
    // these are how the minimum size of the sentence compares to
    //   sum of the size of the words; larger = more spread out
    private static final double VERT_SCALE = 1;
    private static final double HORIZ_SCALE = 1;
    private static final int WORD_MARGIN = 0;

    private Vector words, streams;
    
    private String lang = null;
    private int type, nStreams;
    private boolean streamVisible[];
    private boolean inverseLayout;
    private Probability prob = new Probability();
    private Web web = null;
    private Dimension minSize = null;
    
    /** Creates a new sentence with the given type and axis. */
    public Sentence(int intype, int axis) {
	super(axis);
	type = intype;
	inverseLayout = (type == SOURCE && axis == BoxLayout.X_AXIS);
	//axis = inaxis;
	words = new Vector();
	streams = new Vector();
	streamVisible = null;
	//height = 2 * WORD_MARGIN + NULL_WORD.getUI().getPreferredSize(NULL_WORD).height;
	//addComponentListener(this);
    }
    /** Creates a duplicate of the given sentence on the new axis. */
    public Sentence(Sentence ins, int axis) {
	super(axis);
	//axis = inaxis;
	words = ins.words;
	streams = ins.streams;
	streamVisible = ins.streamVisible;
	inverseLayout = ins.inverseLayout;
	nStreams = ins.nStreams;
	lang = ins.lang;
	type = ins.type;
	prob = new Probability(ins.prob);
	//height = ins.height;
	web = ins.web;
	//addComponentListener(this);
	updateAxis();
    }

    ////////////////// READ METHODS //////////////////

    /** 
     * Reads a sentence probability from an alignment file.
     * @see AlnLoader
     */
    public void readProb(StreamTokenizer st) throws FormatException {
	//prob.read(st, true);
	int tok;
	boolean done = false;
	try{    
	    while (!done) {
		switch(st.nextToken()) {
		case st.TT_NUMBER:
		    throw new FormatException("Zut alors!  Se  01");
		case st.TT_EOF:
		    throw new FormatException("Unexpected end of file.");
		case st.TT_WORD:
		    prob = new Probability(Double.valueOf(st.sval).doubleValue(), true);
		    done = true;
		}
	    }
	} catch (IOException ioe) {
	    throw new FormatException(ioe.getMessage());
	}
    }
    
    /** 
     * Reads a fertility table from an alignment file.
     * @see AlnLoader
     */
    public void readFertility(StreamTokenizer st, boolean corr)
	throws FormatException {
	if (!isTarget())
	    throw new FormatException("Attempt to read fertilities " + 
				      "into non-target sentence.");
	read(st, corr, FERT, null);
    }
    
    /** 
     * Reads a translation table from an alignment file.
     * @see AlnLoader
     */
    public void readTranslations(StreamTokenizer st, boolean corr, 
				 Sentence oppositeSent)
	throws FormatException {
	if (isHuman())
	    throw new FormatException("Attempt to read translations " + 
				      "into human sentence.");
	read(st, corr, TRANS, oppositeSent);
    }
    
    /** 
     * Reads a distortion table from an alignment file.
     * @see AlnLoader
     */
    public void readDistortions(StreamTokenizer st)
	throws FormatException {
	if (!isTarget())
	    throw new FormatException("Attempt to read distortions " + 
				      "into non-target sentence.");
	read(st, true, DIST, null);
    }
    
    /** 
     * Reads a sentence from an alignment file.
     * @see AlnLoader
     */
    public void readWords(StreamTokenizer st) throws FormatException {
	nStreams = 0;
	streamVisible = null;

	read(st, true, WORDS, null);

	if (nStreams > 0)
	    streamVisible = new boolean[nStreams];
	for (int i = 0; i < nStreams; i++)
	    streamVisible[i] = true;
	
	javax.swing.SwingUtilities.invokeLater(new WordAdder(this, true));
    }
    
    private static final int FERT = 0;
    private static final int DIST = 1;
    private static final int TRANS = 2;
    private static final int WORDS = 3;
    
    // generic read function which takes care of all i/o operations.
    private void read(StreamTokenizer st, boolean corr, int which, 
		      Sentence oppositeSent) throws FormatException {
	if (type == SOURCE && (which == FERT || which == DIST))
	    throw new FormatException("Tried to read fertility or " +
				      "distortion into source sentence!");

	boolean done = false;
	Word w;
	
	st.eolIsSignificant(which != WORDS);
	
	try {
	    while (!done) {
		switch (st.nextToken()) {
		case st.TT_EOF:
		    throw new FormatException("Unexpected end of file.");
		case st.TT_WORD:
		    //DBG*/System.err.println("Sentence got token: " + st.sval);
		    if (st.sval.startsWith("</")) {
			done = true;
			break;
		    }
		    switch (which) {
		    case FERT:
			w = getWordAt(Integer.parseInt(st.sval));
			((TargetWord) w).readFertility(st, corr);
			break;
		    case DIST:
			w = getWordAt(Integer.parseInt(st.sval));
			((TargetWord) w).readDistortion(st, corr);
			break;
		    case TRANS:
			w = getWordAt(Integer.parseInt(st.sval));
			w.readTranslation(st, oppositeSent, corr);
			break;
		    case WORDS:
			addWord(st.sval);
			break;
		    }
		    /*if (which != WORDS)
		      throw new FormatException("Unexpected token: " + st.sval);
		      else 
		      addWord(st.sval);*/
		    break;
		case st.TT_NUMBER:  throw new FormatException("Zut alors!  Se  02");
		    //DBG*/System.err.println("Sentence got token: " + st.nval);
		    /*w = getWordAt((int) st.nval);
		      switch (which) {
		      case FERT:
		      ((TargetWord) w).readFertility(st, corr);
		      break;
		      case DIST:
		      ((TargetWord) w).readDistortion(st, corr);
		      break;
		      case TRANS:
		      w.readTranslation(st, oppositeSent, corr);
		      break;
		      case WORDS:
		      addWord(String.valueOf(st.nval));
		      break;
			}*/
		}
	    }
	} catch (IOException ioe) {
	    throw new FormatException(ioe.getMessage());
	} finally {
	    st.eolIsSignificant(true);
	}
    }

    private void addWord(String s) {
	Word w = null;
	Vector v = null;
	StringTokenizer strtok = 
	    new StringTokenizer(s, "|", false);
	int count = strtok.countTokens()-1;
	String title = strtok.nextToken();
	if (count > 0) {
	    StreamLabel tmp[] = new StreamLabel[count];
	    for (int i = 0; i < count; i++) {
		tmp[i] = new StreamLabel(strtok.nextToken());
		tmp[i].setHorizontalTextPosition(JLabel.CENTER);
		tmp[i].setAlignmentX((float) 0.5);
		tmp[i].setAlignmentY((float) 0.5);
	    }
	    streams.add(tmp);
	    if (count > nStreams)
		nStreams = count;
	} else
	    streams.add(null);
	
	switch (type) {
	case SOURCE:
	    w = new SourceWord(title, words.size(), web);
	    break;
	case TARGET:
	    w = new TargetWord(title, words.size(), web);
	    break;
	case HUMAN:
	    w = new HumanWord(title, words.size(), web);
	    break;
	}

	words.addElement(w);
	//*DBG*/System.err.println("Created word with text: " + w.getText());
    }
    /**
     * Update the sentence to reflect the current <code>web.axis</code>.
     */
    public void updateAxis() {
	Word w;
	boolean horiz = (web.getAxis() == BoxLayout.X_AXIS);
	//System.err.println("Updating axis, horiz=" + horiz);
	inverseLayout = (type == SOURCE);
	Enumeration e = words.elements();
	while (e.hasMoreElements()) {
	    w = (Word) e.nextElement();
	    if (horiz) {
		w.setAlignmentX(w.CENTER_ALIGNMENT);
		w.setHorizontalAlignment(w.CENTER);
	    } else {
		if (type == SOURCE) {
		    w.setAlignmentX(w.RIGHT_ALIGNMENT);
		    w.setHorizontalAlignment(w.RIGHT);
		} else {
		    w.setAlignmentX(w.LEFT_ALIGNMENT);
		    w.setHorizontalAlignment(w.LEFT);
		}
	    }
	}
	
	updateStreams();
	updateSize();
    }
    
    /**
     * Recalculate the minimum and preferred sizes.
     */
    private void updateSize() {
	if (words.isEmpty()) {
	    minSize = null;
	    return;
	}

	boolean horiz = (web.getAxis() == BoxLayout.X_AXIS);
	boolean single = true;
	Word w = new Word();
	ComponentUI ui = w.getUI();
	Dimension d = ui.getPreferredSize(w);
	Enumeration e = words.elements();
	
	if (nStreams > 1) {
	    for (int i = 0; i < streamVisible.length; i++)
		if (streamVisible[i])
		    single = false;
	}
	minSize = new Dimension(0, 0);
	
	while (e.hasMoreElements()) {
	    w = (Word) e.nextElement();
	    d = w.getParent().getMinimumSize();//ui.getPreferredSize(w.getParent());
	    if (horiz) {
		minSize.width += d.width;
		if (d.height > minSize.height)
		    minSize.height = d.height;
	    } else {
		minSize.height += d.height;
		if (d.width > minSize.width)
		    minSize.width = d.width;
	    }
	}
	if (horiz) {
	    minSize.width *= HORIZ_SCALE;
	    minSize.height *= VERT_SCALE; //= (int) (d.height * VERT_SCALE);
	} else {
	    minSize.height *= VERT_SCALE; //(int) (d.height * length() * VERT_SCALE);
	    minSize.width *= HORIZ_SCALE;
	}
	//*DBG*/System.err.println(type() + " size: " + minSize);
	//*DBG*/System.err.println(d.toString());
    }
    

    ////////////////// LISTENER METHODS //////////////////
    /*
    public void componentResized(ComponentEvent ce) {
	//updateSize();
    }

    public void componentMoved(ComponentEvent ce) {}
    public void componentHidden(ComponentEvent ce) { }
    public void componentShown(ComponentEvent ce) { }
    */
    ////////////////// ACCESSORS //////////////////
    
    /** Returns a string representation of the sentence. */
    public String toString() {
	StringBuffer out = new StringBuffer("[**" + getTypeString() + " sent: ");
	Enumeration e = words.elements();
	while (e.hasMoreElements())
	    out.append(((Word) e.nextElement()).toString());
	out.append("**]");
	return out.toString();
    }
    /** 
     * Clears all data related to this sentence. 
     */
    public void clear() {
	removeAll();
	words.clear();
	streams.clear();
    }
    /**
     * Displays or hides the given stream and updates the display.
     * @see #isStreamVisible(int)
     */
    public void setStreamVisible(int str, boolean vis) {
	if (streamVisible == null || str >= streamVisible.length ||
	    streamVisible[str] == vis)
	    return;
	streamVisible[str] = vis;
	updateStreams();
	updateSize();
    }
    /**
     * Returns whether the given stream is visible.
     * @see #setStreamVisible(int, boolean)
     */
    public boolean isStreamVisible(int str) {
	if (streamVisible == null || str >= streamVisible.length)
	    return false;
	return streamVisible[str];
    }
    private void updateStreams() {
	// exec wordAdder in this thread
	new WordAdder(this, false).run();
	/*
	JLabel s[];
	Word w;
	Container c;
	int i, j;
	
	for (i = 0; i < streams.size(); i++) {
	    s = (JLabel[]) streams.elementAt(i);
	    if (s == null)
		continue;
	    w = (Word) words.elementAt(i);
	    c = w.getParent();
	    c.removeAll();
	    if (inverseLayout) {
		for (j = s.length-1; j >= 0; j--) {
		    if (streamVisible[j]) {
			c.add(s[j]);
		    }
		}
		c.add(w);
	    } else {
		c.add(w);
		for (j = 0; j < s.length; j++) {
		    if (streamVisible[j])
			c.add(s[j]);
		}
	    }
	}
	*/
	//updateSize();
    }
    /**
     * Returns the type of the sentence.
     * @see #getTypeString()
     * @see #SOURCE
     * @see #TARGET
     * @see #HUMAN
     */
    public int getType() {
	return type;
    }
    /** 
     * Returns a string representing the type of the current sentence.
     * Returns one of "reference", "target", "source", or "unknown".
     * @see #getType()
     */
    public String getTypeString() {
	switch(type) {
	case HUMAN:
	    return "reference";
	case TARGET:
	    return "target";
	case SOURCE:
	    return "source";
	default:
	    return "unknown";
	}
    }
    /**
     * Returns <code>true</code> if this is a source sentence.
     * @see #getType()
     */
    public boolean isSource() {
	return (type == SOURCE);
    }
    /**
     * Returns <code>true</code> if this is a target sentence.
     * @see #getType()
     */
    public boolean isTarget() {
	return (type == TARGET);
    }
    /**
     * Returns <code>true</code> if this is a human sentence.
     * @see #getType()
     */
    public boolean isHuman() {
	return (type == HUMAN);
    }
    /**
     * Sets the Web associated with this sentence to the given value.
     * @see #getWeb()
     */
    public void setWeb(Web w) {
	web = w;
    }
    /**
     * Returns the Web associated with this sentence.
     * @see #setWeb(Web)
     */
    public Web getWeb() {
	return web;
    }
    /**
     * Sets the probability of this sentence to the given value.
     * @see #getProb()
     */
    public void setProb(double p) {
	prob = new Probability(p, true);
    }
    /**
     * Returns the probability of this sentence.
     * @see #setProb(double)
     */
    public double getProb() {
	return prob.val();
    }
    /**
     * Sets the language of this sentence to the given value.
     * @see #getLanguage()
     */
    public void setLanguage(String inl) {
	lang = inl;
    }
    /**
     * Returns the language of this sentence.
     * @see #setLanguage(String)
     */
    public String getLanguage() {
	return lang;
    }
    /**
     * Returns an enumeration of the Words in this sentence.
     * @see Word
     */
    public Enumeration getWords() {
	return words.elements();
    }
    /**
     * Returns the Word at the given index.
     * @see Word
     */
    public Word getWordAt(int index) {
	if (index == -1)
	    return web.getTrans().NULL_WORD;
	return (Word) words.elementAt(index);
    }
    /**
     * Returns the Word at the given index.
     * @see Word
     */
    public JLabel[] getStreamAt(int index) {
	if (index == -1)
	    return null;
	return (JLabel []) streams.elementAt(index);
    }
    /**
     * Returns the number of Words in this sentence.
     */
    public int getLength() {
	return words.size();
    }
    /**
     * Returns the x coordinate of the location of the Word at the given
     * index.
     * @see #getWordYLoc(int)
     */
    public int getWordXLoc(int index) {
	Word w = (Word) words.elementAt(index);
	Container c = (w).getParent();
	return (c.getX() + w.getWidth() / 2);
    }
    /**
     * Returns the y coordinate of the location of the Word at the given
     * index.
     * @see #getWordXLoc(int)
     */
    public int getWordYLoc(int index) {
	Container c = ((Word) words.elementAt(index)).getParent();
	return (c.getY() + c.getHeight() / 2);
    }
    
    /**
     * Overrides Box.getPreferredSize().
     * @see #getMaximumSize()
     * @see #getMinimumSize()
     */
    public Dimension getPreferredSize() {
	if (minSize == null)
	    return super.getPreferredSize();
	
	return minSize;
    }
    /** Finishes initialization of the sentence. */
    public void finish() {
	Enumeration e = words.elements();
	while (e.hasMoreElements())
	    ((Word) e.nextElement()).updateColors();
    }
    
    /**
     * Overrides Box.getMaximumSize().
     * @see #getPreferredSize()
     * @see #getMinimumSize()
     */
    public Dimension getMaximumSize() {
	if (minSize == null)
	    return super.getMaximumSize();
	
	if (web.getAxis() == BoxLayout.X_AXIS)
	    return new Dimension(500000, minSize.height);
	else
	    return new Dimension(minSize.width, 500000);
    }

    /**
     * Overrides Box.getMinimumSize().
     * @see #getMaximumSize()
     * @see #getPreferredSize()
     */
    public Dimension getMinimumSize() {
	if (minSize == null)
	    return super.getMinimumSize();
	
	return minSize;
    }
    
    ////////////////// PRIVATE CLASSES //////////////////
    
    /** 
     * Simple extension of JLabel to represent streams.
     * @author Mike Jahr
     * @see Sentence
     */
    private static class StreamLabel extends JLabel {
	public StreamLabel(String s) {
	    super(s);
	}
    }
    
    /**
     * Thread utility class which adds all words to the given
     * sentence. Required because the loader is not running in the Swing
     * thread, so Sentence has the Swing thread execute this class.
     * @author Mike Jahr
     * @see Sentence
     */
    protected static class WordAdder extends Thread {
	private Sentence s;
	private boolean first;
	static private boolean RUNNING = false;
	
	public WordAdder(Sentence ins, boolean infirst) {
	    super();
	    s = ins;
	}
	
	public void run() {
	    if (RUNNING) return;
	    RUNNING = true;
	    boolean first = true;
	    boolean horiz = (s.web.getAxis() == BoxLayout.X_AXIS);
	    int i, j, len, added;
	    int axis = (horiz) ? BoxLayout.Y_AXIS : BoxLayout.X_AXIS;
	    Box b;
	    JLabel str[];
	    Word w;

	    s.removeAll();
	    
	    //*DBG*/System.err.println(s.words.toString());
	    for (i = 0; i < s.getLength(); i++) {
		w = (Word) s.getWordAt(i);
		//*DBG*/System.err.println("Adding word " + w.getText());
		b = new Box(axis);
		str = (JLabel[]) s.getStreamAt(i);
		len = ((str == null) ? 0 : str.length);

		added = 0;
		if (s.inverseLayout) {
		    for (j = len; j > 0; j--) {
			if (s.isStreamVisible(j-1)) {
			    b.add(str[j-1]);
			    b.add(b.createGlue());
			    added++;
			}
		    }
		    if (added == 0)
			b.add(b.createGlue());
		    b.add(w);
		} else {
		    b.add(w);
		    for (j = 0; j < len; j++) {
			if (s.isStreamVisible(j)) {
			    b.add(b.createGlue());
			    b.add(str[j]);
			    added++;
			}
		    }
		    if (added == 0)
			b.add(b.createGlue());
		}
		
		s.add(s.createGlue());
		s.add(b);
	    }
	    s.add(s.createGlue());

	    if (first) {
		s.updateAxis();
		s.validate();
	    }
	    //DBG*/System.err.println("Added all words.");
	    RUNNING = false;
	}
    }
}
