/* Translation.java
 * ================
 * Mike Jahr, 7/15/99
 * Machine Translation group, WS99
 * Center for Language and Speech Processing
 */

import java.lang.String;
import java.util.Properties;
import java.util.StringTokenizer;
import java.io.StreamTokenizer;
import java.io.IOException;
import javax.swing.table.*;
import javax.swing.JTextArea;
import javax.swing.BoxLayout;
import java.util.Vector;

/**
 * Top-level class in the data structure; keeps track of all information
 * pertaining to a particular sentence pair. Instance variables are public 
 * for easy access. Represented visually as a text area on screen.
 * @author Mike Jahr, Noah A. Smith
 * @see Cairo
 */
public class Translation extends JTextArea {
    /**
     * There is only ever one Null word 
     */
    public final NullWord NULL_WORD = new NullWord();

    /** The {@link Cairo Cairo} object which displays this translation. */
    public Cairo master;
    /** {@link Sentence Sentence} objects for the translation. */
    public Sentence source, target, human;
    /** The {@link Web Web} object. */
    public Web web;
    /** The {@link LanguageModel LanguageModel} object. */
    public LanguageModel lm;
    /** 
     * The current display axis; should always be either BoxLayout.X_AXIS 
     * or BoxLayout.Y_AXIS.
     */
    public int axis;
    /** 
     * The character encoding of the current file; <code>null</code> if
     * the default encoding is used. 
     */
    public String encoding;

    /** 
     * Creates a <code>Translation</code> object with the given axis and
     * the given <code>Cairo</code>. Initializes all of its members.
     */
    public Translation(Cairo c, int axis) {
	source = new Sentence(Sentence.SOURCE, axis);
	target = new Sentence(Sentence.TARGET, axis);
	human = new Sentence(Sentence.HUMAN, axis);
	web = new Web(this, c, axis);
	master = c;
	source.setWeb(web);
	target.setWeb(web);
	human.setWeb(web);
	lm = new LanguageModel(web);
	encoding = null;
	
	setEditable(false);
	setOpaque(false);
	setLineWrap(false);
    }
    /** 
     * Called when all data has been read; finishes initialization of
     * the Sentences and Web.
     * @see Sentence#finish()
     * @see Web#finish()
     */
    public void finish() {
	source.finish();
	target.finish();
	human.finish();
	web.finish();
    }
    /** 
     * Notifies members of the new axis, and updates the <code>axis</code> 
     * member.
     * @see Web#setAxis(int)
     */
    public void setAxis(int newAxis) {
	if (web.getAxis() == newAxis)
	    return;
	
	web.setAxis(newAxis);

	source = new Sentence(source, newAxis);
	target = new Sentence(target, newAxis);
	human = new Sentence(human, newAxis);
    }
    /** 
     * Updates sentences to reflect the visibility of the given stream.
     * @see Sentence#setStreamVisible(int, boolean)
     */
    public void setStreamVisible(int s, boolean vis) {
	source.setStreamVisible(s, vis);
	target.setStreamVisible(s, vis);
	human.setStreamVisible(s, vis);
	web.getParent().validate();
	//web.getParent().repaint();
    }
    
    /** 
     * Clears all information relating to this sentence pair and resets
     * all members.
     * @see LanguageModel#clear()
     * @see Sentence#clear()
     * @see Web#clear()
     */
    public void clear() {
	NULL_WORD.clear();
	master.updateStreams(null);
	lm.clear();
	human.clear();
	source.clear();
	target.clear();
	web.clear();
	setText("");
	encoding = null;
    }
    
    /** 
     * Reads in a header from an alignment file.
     * @see AlnLoader
     */
    /* Read in a header from a formatted alignment file ('st'). */
    public void readHeader(StreamTokenizer st) throws FormatException {
	StringBuffer buff = new StringBuffer("");
	String key, temp;
	
	boolean done = false, secdone = false;
	if(!st.sval.equalsIgnoreCase("<header>"))
	    throw new FormatException("Corrupt file format!  Tr  01");
	try{
	    //lm.clear(); /* when you load a new file, you need to empty the language model */
	    switch(st.nextToken()){
	    case st.TT_EOF:
		throw new FormatException("Unexpected end of file!  Tr  02");
	    case st.TT_WORD:
		throw new FormatException("Expecting end of line," + 
					  " got string!  Tr  03");
	    case st.TT_NUMBER:
		throw new FormatException("Zut alors!  Tr  04");
	    }
	    while(!done){
		switch(st.nextToken()){
		case st.TT_WORD:
		    if(st.sval.equalsIgnoreCase("</header>")){
			done = true;
			break;
		    }
		    if(st.sval.equalsIgnoreCase("date:"))
			key = "Date:";
		    else if(st.sval.equalsIgnoreCase("source:"))
			key = "Source:";
		    else if(st.sval.equalsIgnoreCase("target:"))
			key = "Target:";
		    else if(st.sval.equalsIgnoreCase("model:"))
			key = "Model:";
		    else if(st.sval.equalsIgnoreCase("corpus:"))
			key = "Corpus:";
		    else if(st.sval.equalsIgnoreCase("source_sentence_id:"))
			key = "Source ID:";
		    else if(st.sval.equalsIgnoreCase("target_sentence_id:"))
			key = "Target ID:";
		    else if(st.sval.equalsIgnoreCase("lm_ngram_length:"))
			key = "Language Model N-gram length:";
		    else if(st.sval.equalsIgnoreCase("streams:"))
			key = "Streams:";
		    else if(st.sval.equalsIgnoreCase("encoding:"))
			key = "Encoding:";
		    else
			key = st.sval;
		    temp = "";
		    secdone = false;
		    while(!secdone){
			switch(st.nextToken()){
			case st.TT_WORD:
			    temp = temp + st.sval + " ";
			    break;
			case st.TT_NUMBER:  throw new FormatException("Zut alors!  Tr  04.1");
			case st.TT_EOF:
			    throw new FormatException("Unexpected end of file!"
						      + " Tr  04.01");
			case st.TT_EOL:
			    secdone = true;
			    st.pushBack();
			}
		    }
		    
		    temp = temp.trim();
		    
		    if(!key.equals("Streams:"))
			buff.append("\n" + key + "  " + temp);
		    else{
			master.updateStreams(temp);
			/*
			StringTokenizer stringTok = new StringTokenizer(temp, "|", false);
			streamNames = new String[stringTok.countTokens()];
			for(int i = 0; i < stringTok.countTokens(); i++)
			    streamNames[i] = stringTok.nextToken();
			*/
		    }
		    //attribNames.add(key);
		    //attribVals.add(temp);
		    if (key.equals("Source:")) {
			source.setLanguage(temp);
		    } else if (key.equals("Target:")) {
			target.setLanguage(temp);
			human.setLanguage(temp);
		    } else if(key.equals("Language Model N-gram length:")){
			//System.err.println(temp);
			lm.setGram(Integer.parseInt(temp.trim()));
		    } else if(key.equals("Encoding:")) {
			encoding = temp;
		    }			
		    break;
		default:
		    throw new 
			FormatException("Expecting attribute!  Tr  05");
		}
		switch(st.nextToken()){
		case st.TT_EOF:  
		    throw new FormatException("Unexpected end of file!  Tr  06");
		case st.TT_WORD:  
		    throw new FormatException("Expecting end of line, got string!  Tr  07");
		case st.TT_NUMBER:  
		    throw new FormatException("Zut alors!  Tr  08");
		}
	    }
	} catch(IOException ioe) {
	    throw new FormatException("Unexpected end of file!  Tr  09");
	}
	buff.deleteCharAt(0);
	setText(buff.toString());
    }
}
